from nostr_bot.bot import RegistrationBot
import os

class ShibbolethRegistrationBot(RegistrationBot):
    SHIBBOLETH = os.getenv("SHIBBOLETH")

    async def is_valid(self, event):
        return event.content == self.SHIBBOLETH


class PaymentRegistrationBot(RegistrationBot):
    AMOUNT = os.getenv("AMOUNT")

    async def is_valid(self, event):
        return event.content == 'here is {self.AMOUNT}'




"""
TattleBot will listen for kind=1984 reports and notify the subject of the report that they've been tattled on.

You can subclass TattleBot and override handle_message(event, tattle_subject, message) to do something more interesting with the generated message
Override create_message(event, report_type, tattled_event_id, impersonation) to generate a different response

To run:
TATTLE_WATCH=<mypubkey> NOSTR_KEY=<bot_private_key> nostr-bot run -c nostr_bot.examples.reporting.TattleBot -r wss://my.relay.biz

"""

from nostr_bot.bot import CommunicatorBot
from aionostr.util import to_nip19
import shelve
import os
import time


class TattleBot(CommunicatorBot):
    SEND_MESSAGE = False

    shelf = shelve.open('tattlebot')

    def get_query(self):
        query = {
            'kinds': [1984]
        }
        last_seen = self.get_last_seen()
        if last_seen:
            query['since'] = last_seen
        pubkeys = self.get_watch_for_pubkeys()
        if pubkeys:
            query['#p'] = pubkeys
        return query

    def get_last_seen(self):
        return self.shelf.get('last_seen', 0)

    def set_last_seen(self, event):
        self.shelf['last_seen'] = max(event.created_at, self.get_last_seen())

    def get_watch_for_pubkeys(self):
        pubkeys = os.getenv('TATTLE_WATCH', '').split(',')
        if all(pubkeys):
            return pubkeys

    async def handle_event(self, event):
        if event.id in self.shelf:
            self.log.info("Skipping %s", event.id)
            return
        report_type = ''
        tattled_event_id = ''
        tattle_subject = ''
        impersonation = ''
        for tag in event.tags:
            if tag[0] == 'report':
                report_type = tag[1]
            elif tag[0] == 'e':
                tattled_event_id = tag[1]
            elif tag[0] == 'p':
                if len(tag) == 3 and tag[2] == 'impersonation':
                    impersonation = tag[1]
                else:
                    tattle_subject = tag[1]

        message = self.create_message(event, report_type, tattled_event_id, impersonation)

        response = await self.handle_message(event, tattle_subject, message)
        self.set_last_seen(event)
        self.shelf[event.id] = {'seen': time.time(), 'response': response}

    def create_message(self, event, report_type, tattled_event_id, impersonation):
        reporter = to_nip19('npub', event.pubkey)
        if tattled_event_id:
            tattle_note = to_nip19('note', tattled_event_id)
        else:
            tattle_note = '<somewhere>'
        if report_type == 'spam':
            reason = 'spamming'
        elif report_type == 'illegal':
            reason = 'doing something illegal'
        elif report_type == 'impersonation':
            reason = f'impersonating {to_nip19("npub", impersonation)}'
        else:
            reason = report_type
        response = f'''YOU'RE IN TROUBLE
{reporter} tattled on you for {reason} in {tattle_note}.
They said you were "{event.content}"

Just letting you know.

Sent by TattleBot.
'''
        return response

    async def handle_message(self, event, tattle_subject, message):
        if tattle_subject and message:
            if self.SEND_MESSAGE:
                dm = self.make_dm(tattle_subject, content=message)
                self.log.debug(str(dm))
                await self.reply(dm)
                self.log.info("Alerted %s about tattling on %s with dm %s", tattle_subject, event.id, dm.id)
                return dm.id
            else:
                self.log.info("%s tattled on %s. Sending:\n%s", event.pubkey, tattle_subject, message)



"""
New User bot

This will listen for kind=0 events and send a dm to itself, welcoming the user

"""
from nostr_bot import CommunicatorBot
import json


class WelcomeBot(CommunicatorBot):
    LISTEN_KIND = 0
    LISTEN_PUBKEY = None

    async def handle_event(self, event):
        meta = json.loads(event.content)
        name = meta.get('display_name', '') or meta.get('name', '') or event.pubkey
        dm = self.make_dm(self.PUBLIC_KEY, content=f"Welcome, {name}!")
        self.log.info("Welcoming %s with %s", name, dm.id)
        await self.reply(dm)



import logging
import os

from aionostr.relay import Manager
from aionostr.event import Event, loads, dumps

__all__ = ('NostrBot', 'RPCBot', 'CommunicatorBot')


class NostrBot:
    """
    Base bot class that listens for a query and logs the results

    Override `handle_event(event)` to do something interesting
    """
    LISTEN_KIND = 1
    LISTEN_PUBKEY = ''
    LIMIT = 1
    RELAYS = ['ws://localhost:6969']
    PRIVATE_KEY = os.getenv('NOSTR_KEY', 'd3b7207018ac76dfab82100a6c07a42c68b8efc4898b96d2882b8a7636dd0498')

    def __init__(self):
        self.log = logging.getLogger(self.get_origin())
        self._manager = None

    def get_origin(self):
        return self.__class__.__name__

    def get_manager(self):
        pk = self.private_key
        if pk:
            pk = pk.hex()
        return Manager(self.get_relays(), origin=self.get_origin(), private_key=pk)

    def get_relays(self):
        return self.RELAYS

    def get_query(self):
        filter_obj = {
            'limit': self.LIMIT
        }
        if self.LISTEN_PUBKEY:
            filter_obj['authors'] = [self.LISTEN_PUBKEY]
        if self.LISTEN_KIND is not None:
            filter_obj['kinds'] = [self.LISTEN_KIND]
        return filter_obj

    @property
    def private_key(self):
        if not self.PRIVATE_KEY:
            return None
        from aionostr.key import PrivateKey
        from aionostr.util import from_nip19
        if self.PRIVATE_KEY.startswith('nsec'):
            pk = from_nip19(self.PRIVATE_KEY)['object']
        else:
            pk = PrivateKey(bytes.fromhex(self.PRIVATE_KEY))
        return pk

    @property
    def manager(self):
        if not self._manager:
            self._manager = self.get_manager()
        return self._manager

    @manager.setter
    def manager(self, manager):
        self._manager = manager

    async def start(self):
        if self.private_key:
            self.manager.private_key = self.private_key.hex()
        await self.manager.connect()

        query = self.get_query()
        self.log.info("Running query %s on %s", query, self.get_relays())


        async for event in self.manager.get_events(query, only_stored=False):
            try:
                if not event.verify():
                    self.log.warning('Invalid event: %s', event.id)
                    continue
            except Exception as e:
                self.log.error(str(e))
                continue
            try:
                await self.handle_event(event)
            except Exception:
                self.log.exception('handle_event')

    async def handle_event(self, event: Event):
        """
        Override this to do useful things
        """
        self.log.info(str(event))


class CommunicatorBot(NostrBot):
    """
    A bot that can make DM's and reply
    """
    PUBLIC_KEY = ''
    
    def __init__(self):
        super().__init__()
        if not self.PUBLIC_KEY:
            self.PUBLIC_KEY = self.private_key.public_key.hex()
        elif self.PUBLIC_KEY.startswith('npub'):
            from aionostr.util import from_nip19
            self.PUBLIC_KEY = from_nip19(self.PUBLIC_KEY)['object'].hex()

    def make_event(self, encrypt_to=None, **event_args):
        if encrypt_to:
            event_args['content'] = self.private_key.encrypt_message(event_args['content'], encrypt_to)
        if not event_args.get('pubkey'):
            event_args['pubkey'] = self.PUBLIC_KEY
        event = Event(**event_args)

        event.sign(self.private_key.hex())
        return event

    def make_dm(self, encrypt_to, **event_args):
        tags = event_args.get('tags', [])
        tags.append(["p", encrypt_to])
        event_args['kind'] = 4
        event_args['tags'] = tags
        event = self.make_event(encrypt_to=encrypt_to, **event_args)
        return event

    async def reply(self, event):
        self.log.debug("Replying with %s", event)
        await self.manager.add_event(event, check_response=True)


class RPCBot(CommunicatorBot):
    """
    A bot that listens for (optionally encrypted) messages in the form:
    {
        "method": "name",
        "args": ["arg1", 1, "arg3"]
    }

    By default, these are kind 22222 -- ephemeral events
    """
    LISTEN_KIND = 22222
    ENCRYPTED = True

    def __init__(self):
        super().__init__()
        self.LISTEN_PUBKEY = self.PUBLIC_KEY

    async def handle_event(self, event: Event):
        content = event.content
        if self.ENCRYPTED:
            try:
                content = self.private_key.decrypt_message(content, event.pubkey)
            except Exception:
                self.log.exception("decrypt")
                return
        try:
            command = loads(content)
        except Exception:
            self.log.exception("json")
            return
        try:
            func = getattr(self, f"on_{command['method']}")
        except (KeyError, AttributeError, TypeError):
            self.log.error("%s", command)
            return
        try:
            args = command['args']
            response = await func(event, *args)
            self.log.debug("Command %s from %s", command, event)
        except Exception as e:
            self.log.exception(str(e))
        else:
            if response and 'event' in response:
                await self.reply(response['event'])

    def make_response(self, event, **kwargs):
        if self.ENCRYPTED:
            kwargs['encrypt_to'] = event.pubkey
        response = {
            'event': self.make_event(**kwargs)
        }
        return response


class RegistrationBot(CommunicatorBot):
    LISTEN_KIND = 11141
    LIMIT = 100

    async def handle_event(self, event: Event):
        if not event.verify():
            self.log("Bad event: %s", event.id)
            return
        self.log.info("Got registration request %s", event)

        if await self.is_valid(event):
            self.log.info("Valid registration from %s", event.pubkey)
            if await self.register(event):
                self.log.info("Registered pubkey %s", event.pubkey)

    async def is_valid(self, event: Event):
        valid = False
        for tag in event.tags:
            if tag[0] == 'relay':
                if tag[1] in self.RELAYS:
                    valid = True
        return valid

    async def register(self, event: Event):
        return False


async def start_multiple(bots, relays=None):
    """
    Start multiple bots in their own task
    """
    import asyncio

    if relays:
        first_manager = None
        for bot in bots:
            bot.RELAYS = relays
            if not first_manager:
                first_manager = bot.manager
            bot.manager = first_manager

    tasks = [asyncio.create_task(bot.start()) for bot in bots]
    try:
        await asyncio.wait(tasks)
    except asyncio.exceptions.CancelledError:
        return



"""Console script for nostr_bot."""
import sys
import click
import asyncio
import os
from functools import wraps
try:
    import uvloop
    uvloop.install()
except ImportError:
    pass


DEFAULT_RELAYS = os.getenv('NOSTR_RELAYS', 'wss://nostr.mom,wss://relay.snort.social').split(',')

def stop():
    loop = asyncio.get_running_loop()
    try:
        loop.stop()
    except RuntimeError:
        return


def async_cmd(func):
  @wraps(func)
  def wrapper(*args, **kwargs):
        async def _run(coro):
            from signal import SIGINT, SIGTERM

            loop = asyncio.get_running_loop()
            for signal_enum in [SIGINT, SIGTERM]:
                loop.add_signal_handler(signal_enum, stop)

            await coro
        coro = func(*args, **kwargs)
        try:
            asyncio.run(_run(coro))
        except RuntimeError:
            return
  return wrapper


@click.group()
def main(args=None):
    """Console script for nostr_bot."""
    return 0


@main.command()
@click.option('-c', '--cls', multiple=True, help='bot class(es) to run', default=['nostr_bot.NostrBot'])
@click.option('-r', 'relays', multiple=True, help='Relay address (can be added multiple times)', default=DEFAULT_RELAYS)
@click.option('-v', '--verbose', help='verbose results', is_flag=True, default=False)
@async_cmd
async def run(relays, cls, verbose):
    """
    Run a bot
    """
    import logging
    import importlib
    from .bot import start_multiple
    logging.basicConfig(format='%(asctime)s %(name)s %(levelname)s – %(message)s', level=logging.DEBUG if verbose else logging.INFO)

    bots = []
    for classname in cls:
        try:
            bot_module, bot_class = classname.rsplit('.', 1)
            bot_class = getattr(importlib.import_module(bot_module), bot_class)
        except (ImportError, AttributeError):
            click.echo(f"Class {classname} not found")
            return -1
        bots.append(bot_class())
    await start_multiple(bots, relays=relays)




if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""
Announce when you receive a DM

Set environment variable PUBLIC_KEY to your public key

Set env var KINDS to a comma-separated list of kinds to listen for

This works on MacOS by calling the 'say' command

You can set env var VOICE to use a different voice
"""
from nostr_bot import NostrBot
import os
import subprocess
import json


class GotMailBot(NostrBot):
    VOICE = os.getenv("VOICE", "Fred")
    KINDS = [int(k) for k in os.getenv("KINDS", "4").split(',')]
    seen_pubkeys = {}

    def get_query(self):
        my_pubkey = os.getenv('PUBLIC_KEY')
        assert my_pubkey, "Set PUBLIC_KEY environment variable"
        return {
            'limit': 1,
            '#p': [my_pubkey],
            'kinds': self.KINDS,
        }

    async def handle_event(self, event):
        if event.pubkey not in self.seen_pubkeys:
            self.seen_pubkeys[event.pubkey] = event.pubkey[-4:]
            async for profile in self.manager.get_events({"authors": [event.pubkey], "kinds": [0]}, single_event=True):
                meta = json.loads(profile.content)
                self.seen_pubkeys[event.pubkey] = meta.get('display_name', '') or meta.get('name', '') or event.pubkey[-4:]

        name = self.seen_pubkeys[event.pubkey]
        kind = "mail"
        if event.kind == 7:
            kind = "a reaction"
        elif event.kind == 1:
            kind = "a reply"
        announcement = f"You've got {kind} from {name}"
        self.log.info(announcement)
        command = ["say", announcement, f"-v", self.VOICE]
        try:
            subprocess.run(command)
        except FileNotFoundError:
            self.log.error("Cannot speak the announcement")



"""
MirrorBot will mirror events from authors you follow. 
The source relays are set with environment variable NOSTR_RELAYS or with -r on the command line

To run:

TARGET=ws://target-relay.biz PUBLIC_KEY=<YOURPUBKEY> nostr-bot run -c nostr_bot.examples.mirror.MirrorFollowersBot
"""
import os
import time
import shelve
from nostr_bot.bot import NostrBot
from aionostr import Manager


class MirrorFollowersBot(NostrBot):
    MY_PUBKEY = os.getenv('PUBLIC_KEY')
    TARGET_RELAY = os.getenv('TARGET')
    shelf = shelve.open(f'mirrorbot-{TARGET_RELAY.replace("://", "-")}')

    def get_query(self):
        return self.query

    def get_last_seen(self):
        return self.shelf.get('last_seen', 1)

    def set_last_seen(self, timestamp):
        if timestamp > self.get_last_seen():
            self.shelf['last_seen'] = timestamp

    async def get_following(self):
        find_query = {
            'kinds': [3],
            'authors': [self.MY_PUBKEY]
        }
        following = [self.MY_PUBKEY]
        self.log.info("Getting following for %s %s", self.MY_PUBKEY, find_query)
        async for event in self.manager.get_events(find_query):
            for tag in event.tags:
                if tag[0] == 'p':
                    following.append(tag[1])
        return following

    async def start(self):
        self.target_manager = Manager([self.TARGET_RELAY])
        await self.target_manager.connect()
        await self.manager.connect()
        self.query = {
            'authors': await self.get_following(),
            'since': self.get_last_seen()
        }
        return await super().start()

    async def handle_event(self, event):
        if event.id in self.shelf:
            return
        await self.target_manager.add_event(event)
        self.log.info("Mirrored %s from %s to %s", event.id[:8], event.pubkey, self.TARGET_RELAY)
        self.set_last_seen(event.created_at)
        self.shelf[event.id] = time.time()



"""
Ping bot
"""

import os
import json
from nostr_bot import RPCBot


class PingBot(RPCBot):
    ENCRYPTED = False

    def __init__(self):
        super().__init__()
        self.LISTEN_PUBKEY = os.getenv("PUBLIC_KEY")

    async def on_ping(self, event, *args):
        self.log.info("Got ping %s", event)
        self.log.info("Sending pong")
        return self.make_response(event, kind=event.kind, content=json.dumps({"method": "pong", "args": []}))

    async def on_pong(self, event, *args):
        self.log.info("Got pong %s", event)